from chisubmit.backend.webapp.api import db
from chisubmit.backend.webapp.api.teams.models import Team, StudentsTeams, AssignmentsTeams
from chisubmit.backend.webapp.api.blueprints import api_endpoint
from flask import jsonify, request, abort
from chisubmit.backend.webapp.api.teams.forms import UpdateTeamInput,\
    CreateTeamInput, UpdateAssignmentTeamInput
from chisubmit.backend.webapp.auth.token import require_apikey
from chisubmit.backend.webapp.auth.authz import check_course_access_or_abort,\
    check_team_access_or_abort
from flask import g
from chisubmit.backend.webapp.api.courses.models import Course
from chisubmit.backend.webapp.api.teams.models import Grade
from chisubmit.backend.webapp.api.types import update_options

@api_endpoint.route('/courses/<course_id>/teams', methods=['GET', 'POST'])
@require_apikey
def teams(course_id):
    course = Course.query.filter_by(id=course_id).first()
    
    if course is None:
        abort(404)    

    check_course_access_or_abort(g.user, course, 404)
    
    if request.method == 'GET':
        # TODO: SQLAlchemy-fy this
        teams = Team.query.filter_by(course_id=course_id).all()

        if not g.user.has_instructor_or_grader_permissions(course):
            teams = [t for t in teams if g.user in t.students]
            
        teams_dict = []
        
        for team in teams:
            extension_policy = course.options.get("extension-policy", None)
            t = team.to_dict()
            t["extensions_available"] = team.get_extensions_available(extension_policy)
            teams_dict.append(t)

        return jsonify(teams=teams_dict)

    check_course_access_or_abort(g.user, course, 404, roles = ["instructor"])

    input_data = request.get_json(force=True)
    if not isinstance(input_data, dict):
        return jsonify(error='Request data must be a JSON Object'), 400

    form = CreateTeamInput.from_json(input_data)
    if not form.validate():
        return jsonify(errors=form.errors), 400

    team = Team()
    form.populate_obj(team)
    db.session.add(team)
    db.session.commit()

    return jsonify({'team': team.to_dict()}), 201


@api_endpoint.route('/courses/<course_id>/teams/<team_id>', methods=['GET', 'PUT'])
@require_apikey
def team(course_id, team_id):
    course = Course.query.filter_by(id=course_id).first()
    
    if course is None:
        abort(404)
            
    team = Team.query.filter_by(id=team_id).first()
    if team is None:
        abort(404)

    check_team_access_or_abort(g.user, team, 404)
    
    if request.method == 'PUT':
        check_team_access_or_abort(g.user, team, 404, roles = ["instructor"])
        input_data = request.get_json(force=True)
        if not isinstance(input_data, dict):
            return jsonify(error='Request data must be a JSON Object'), 400
        form = UpdateTeamInput.from_json(input_data)
        if not form.validate():
            return jsonify(errors=form.errors), 400

        team.set_columns(**form.patch_data)

        if 'students' in form:
            for child_data in form.students.add:
                new_child = StudentsTeams()
                child_data.populate_obj(type("", (), dict(
                    new_child=new_child))(), 'new_child')
                db.session.add(new_child)

        if 'assignments' in form:
            for child_data in form.assignments.add:
                new_child = AssignmentsTeams()
                child_data.populate_obj(type("", (), dict(
                    new_child=new_child))(), 'new_child')
                db.session.add(new_child)

            for child_data in form.assignments.update:
                assignment_id = child_data["assignment_id"].data

                at = AssignmentsTeams.from_id(course_id, team_id, assignment_id)

                if at is None:
                    error_msgs = ["Team %s is not registered for assignment %s" % (team_id, assignment_id)]
                    return jsonify(errors={"grades": error_msgs}), 400

                at.set_columns(**child_data.patch_data)

        if 'grades' in form:
            for child_data in form.grades.add:
                # Does the grade already exist?
                assignment_id = child_data["assignment_id"].data
                grade_component_id = child_data["grade_component_id"].data
                
                at = AssignmentsTeams.from_id(course_id, team_id, assignment_id)
                
                if at is None:
                    error_msgs = ["Team %s is not registered for assignment %s" % (team_id, assignment_id)]
                    return jsonify(errors={"grades": error_msgs}), 400
                
                grade = at.get_grade(grade_component_id)
                
                if grade is None:
                    new_child = Grade()
                    child_data.populate_obj(type("", (), dict(
                        new_child=new_child))(), 'new_child')
                    new_child.course_id = course_id
                    new_child.team_id = team_id                    
                    db.session.add(new_child)
                else:
                    grade.points = child_data["points"].data
                    db.session.add(grade)
                    
            if len(form.grades.penalties) > 0:
                penalties = form.grades.penalties.data
                for penalty in penalties:
                    assignment_id = penalty["assignment_id"]
                    penalty_value = penalty["penalties"]

                    at = AssignmentsTeams.from_id(course_id, team_id, assignment_id)
                    
                    if at is None:
                        error_msgs = ["Team %s is not registered for assignment %s" % (team_id, assignment_id)]
                        return jsonify(errors={"grades": error_msgs}), 400
                                        
                    at.penalties = penalty_value
                    db.session.add(at)
                    
        if 'extras' in form:
            if len(form.extras) > 0:
                update_options(form.extras, team.extras)
                db.session.add(team)                    

        db.session.commit()

    
    extension_policy = course.options.get("extension-policy", None)
    t = team.to_dict()
    t["extensions_available"] = team.get_extensions_available(extension_policy)

    return jsonify({'team': t})




from chisubmit.backend.webapp.api import db
from chisubmit.backend.webapp.api.models.mixins import UniqueModel
from chisubmit.backend.webapp.api.models.json import Serializable


class User(UniqueModel, Serializable, db.Model):
    __tablename__ = 'users'
    id = db.Column(db.Unicode, primary_key=True)
    first_name = db.Column(db.Unicode)
    last_name = db.Column(db.Unicode)
    email = db.Column(db.Unicode)
    api_key = db.Column(db.Unicode)
    admin = db.Column('admin', db.Boolean, server_default='0', nullable=False)
    default_fields = ['id', 'first_name', 'last_name', 'email']
    readonly_fields = ['id']

    @classmethod
    def unique_hash(cls, **kws):
        return kws['id']

    @classmethod
    def unique_filter(cls, query, **kws):
        return query.filter(User.id == kws['id'])
    
    @staticmethod
    def from_id(user_id):
        return User.query.filter_by(id=user_id).first()    
    
    def is_instructor_in(self, course):
        return self in course.instructors        

    def is_student_in(self, course):
        return self in course.students        

    def is_grader_in(self, course):
        return self in course.graders        

    def has_instructor_or_grader_permissions(self, course):
        return self.is_instructor_in(course) or self.is_grader_in(course) or self.admin

    def is_in_course(self, course):
        return self.is_instructor_in(course) or self.is_student_in(course) or self.is_grader_in(course)
    



from chisubmit.backend.webapp.api import db
from chisubmit.backend.webapp.api.models.json import Serializable
import math
from chisubmit.backend.webapp.api.types import UTCDateTime

class Assignment(Serializable):
    __tablename__ = 'assignments'
    id = db.Column(db.Unicode, primary_key=True)
    course_id = db.Column('course_id', db.Integer, db.ForeignKey('courses.id'), primary_key=True)

    name = db.Column(db.Unicode)
    deadline = db.Column(UTCDateTime)
    grade_components = db.relationship("GradeComponent", backref="assignment")
    
    default_fields = ['id', 'name', 'deadline', 'course_id', 
                      'grade_components']
    readonly_fields = ['id', 'grade_components', 'course_id']
    
    @staticmethod
    def from_id(course_id, assignment_id):
        return Assignment.query.filter_by(id=assignment_id, course_id=course_id).first()    
    
        
class GradeComponent(Serializable, db.Model):
    __tablename__ = 'grade_components'
    id = db.Column(db.Unicode, primary_key=True)
    course_id = db.Column('course_id', db.Integer, primary_key=True)
    assignment_id = db.Column('assignment_id', db.Integer, primary_key=True)

    order = db.Column(db.Integer)
    description = db.Column(db.Unicode)
    points = db.Column('points', db.Float)
    
    default_fields = ['id', 'description', 'order', 'points', 'assignment_id']
    readonly_fields = ['id', 'course_id', 'assignment_id']
    
    __table_args__ = (db.ForeignKeyConstraint([assignment_id, course_id],
                                              [Assignment.id, Assignment.course_id]),
                      {})



import click
from chisubmit.cli.common import pass_course
from chisubmit.common import CHISUBMIT_FAIL, CHISUBMIT_SUCCESS,\
    ChisubmitException
import operator
from chisubmit.common.utils import convert_datetime_to_local

@click.command(name="list")
@click.option('--ids', is_flag=True)
@click.option('--assignment', type=str)
@click.option('--include-inactive', is_flag=True)
@pass_course
@click.pass_context
def shared_team_list(ctx, course, ids, assignment, include_inactive):
    teams = course.get_teams()
    teams.sort(key=operator.attrgetter("id"))

    assignment_id = assignment
    assignment = None
    if assignment_id is not None:
        assignment = course.get_assignment(assignment_id)
        if assignment is None:
            print "Assignment %s does not exist" % assignment_id
            ctx.exit(CHISUBMIT_FAIL)

    for team in teams:
        if assignment is not None:
            if not team.has_assignment(assignment.id):
                continue
            
        if not (team.active or include_inactive):
            continue
        
        if ids:
            print team.id
        else:
            if len(team.students) == 0:
                students = "No students"
            else:
                students = "Students: " + ",".join([s.user.id for s in team.students])
            if len(team.assignments) == 0:
                assignments = "No assignments"
            else:
                assignments = "Assignments: " + ",".join([a.assignment_id for a in team.assignments])
            fields = [team.id, students, assignments]

            print "  ".join(fields)

    return CHISUBMIT_SUCCESS

@click.command(name="show")
@click.argument('team_id', type=str)
@pass_course
@click.pass_context
def shared_team_show(ctx, course, team_id):
    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)
        
    print "Team name: %s" % team.id
    print
    print "Extensions available: %i" % team.extensions_available 
    print
    
    if len(team.extras) > 0:
        print "Attributes:"
        for k,v in team.extras.items():
            print " - %s: %s" % (k,v)
        print
    
    if len(team.students) == 0:
        print "No students in this team"
    else:
        print "STUDENTS"
        print "--------"
        for student in team.students:
            if student.status == 0:
                status = "UNCONFIRMED"
            elif student.status == 1:
                status = "CONFIRMED"
            else:
                raise ChisubmitException("Student '%s' in team '%s' has unknown status %i" % (student.user.id, team.id, student.status))
            print "%s: %s, %s  (%s)" % (student.user.id, student.user.last_name, student.user.first_name, status)

    print

    if len(team.assignments) == 0:
        print "This team is not registered for any assignments."
    else:
        print "ASSIGNMENTS"
        print "-----------"
        for ta in team.assignments:
            assignment = course.get_assignment(ta.assignment_id)
            print "ID: %s" % assignment.id
            print "Name: %s" % assignment.name
            print "Deadline: %s" % convert_datetime_to_local(assignment.deadline).isoformat(" ")        
            if ta.submitted_at is not None:
                print "Last submitted at: %s" % convert_datetime_to_local(ta.submitted_at).isoformat(" ")
                print "Commit SHA: %s" % ta.commit_sha
                print "Extensions used: %i" % ta.extensions_used
            else:
                print "NOT SUBMITTED"    
            print    




#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

from chisubmit.client import ApiObject
from chisubmit.client import session
import json
import base64
from chisubmit.common import ChisubmitException

class User(ApiObject):

    _api_attrs = ('id', 'first_name', 'last_name', 'email')
    _primary_key = 'id'
    _singularize = 'user'
    _pluralize = 'users'
           
    # TODO 18DEC14: handle revoking and/or replacing the token
    @classmethod
    def get_token(cls, username, password, reset=False):
        credentials = base64.b64encode(username + ":" + password)
        
        headers={'Authorization': 'Basic ' + credentials}        
        
        data = json.dumps({"reset":reset})
        response = session.post('auth', data=data, headers=headers)
        
        if not response.has_key("api_key") or response["api_key"] is None:
            raise ChisubmitException("Chisubmit server did not return valid credentials")
        else:
            return (response["api_key"], response["exists_prior"], response["is_new"])



from __future__ import unicode_literals



def gen_users(ninstructors, ngraders, nstudents):
    
    def __gen_users(prefix, n):
        u = {}
        for i in range(n):
            username = prefix + str(i+1)
            u[username] = {"first_name": "F_" + username,
                            "last_name": "L_" + username,
                            "id": username,
                            "api_key": username}
        return u
    
    users = {}
    
    users.update(__gen_users("instructor", ninstructors))
    users.update(__gen_users("grader", ngraders))
    users.update(__gen_users("student", nstudents))
    
    return users

users_and_courses = \
{ "users": gen_users(ninstructors = 2, ngraders = 4, nstudents = 8),
 "courses": { "cmsc40100": {"id": "cmsc40100",
                            "name": "Introduction to Software Testing",
                            "instructors": ["instructor1"],
                            "graders": ["grader1", "grader2"],
                            "students": ["student1", "student2", "student3", "student4"],
                            },
              "cmsc40110": {"id": "cmsc40110",
                            "name": "Advanced Software Testing",
                            "instructors": ["instructor2"],
                            "graders": ["grader3", "grader4"],
                            "students": ["student5", "student6", "student7", "student8"]
                           }
            }
}

complete_course = \
{ "users": gen_users(ninstructors = 1, ngraders = 2, nstudents = 4),
 "courses": { "cmsc40100": {"id": "cmsc40100",
                            "name": "Introduction to Software Testing",
                            "instructors": ["instructor1"],
                            "graders": ["grader1", "grader2"],
                            "students": ["student1", "student2", "student3", "student4"],
                            "assignments": { "pa1": {"id": "pa1",
                                                     "name": "Programming Assignment 1",
                                                     "deadline": "2042-01-21T20:00+00:00"},
                                             "pa2": {"id": "pa2",
                                                     "name": "Programming Assignment 2",
                                                     "deadline": "2042-01-28T20:00+00:00"}
                                            }
                            }
            }
}

fixtures = { "users_and_courses": (users_and_courses, 
                                   """Fixture with two users and two courses.
                                   Each course has one of the users as the sole instructor."""),
             "complete_course": (complete_course,
                                 """Fixture with a single course with 1 instructor, 2 graders,
                                 4 students, 2 assignments, but no teams. Represents a course
                                 at the start of the term.""")            
            }


import click
from chisubmit.cli.common import pass_course
from chisubmit.cli.shared.course import shared_course_list,\
    shared_course_set_default, shared_course_get_git_credentials

@click.group(name="course")
@click.pass_context
def student_course(ctx):
    pass




@click.command(name="set-git-username")
@click.argument('username', type=str)
@pass_course
@click.pass_context
def student_course_set_git_username(ctx, course, username):
    course.set_student_repo_option(None, "git-username", username)

student_course.add_command(shared_course_list)
student_course.add_command(shared_course_set_default)
student_course.add_command(student_course_set_git_username)
student_course.add_command(shared_course_get_git_credentials)




#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import os.path
from pkg_resources import Requirement, resource_filename
from chisubmit.common import ChisubmitException
import shutil
import yaml
import collections

DEFAULT_CHISUBMIT_DIR = os.path.expanduser("~/.chisubmit/")
DEFAULT_CONFIG_FILENAME = "chisubmit.conf"

# Based on http://stackoverflow.com/questions/3387691/python-how-to-perfectly-override-a-dict
class Config(collections.MutableMapping):

    IMPLICIT_FIELDS = ["directory"]
    REQUIRED_FIELDS = ["api-url"]
    OPTIONAL_FIELDS = ["default-course", "git-credentials", "server", "api-key"]
    ALL_FIELDS = REQUIRED_FIELDS + OPTIONAL_FIELDS

    def __getitem__(self, key):
        if key in self.options:
            return self.options[key]
        elif key in self.OPTIONAL_FIELDS:
            return None
        else:
            raise KeyError

    def __setitem__(self, key, value):
        if key in self.ALL_FIELDS:
            self.options[key] = value
            self.save()
        else:
            raise KeyError

    def __delitem__(self, key):
        del self.options[key]
        self.save()

    def __iter__(self):
        return iter(self.options)

    def __len__(self):
        return len(self.options)

    def __init__(self, config_dir=None, config_file=None):

        if config_dir is None:
            config_dir = DEFAULT_CHISUBMIT_DIR
        else:
            config_dir = os.path.expanduser(config_dir)

        if config_file is None:
            config_file = config_dir + DEFAULT_CONFIG_FILENAME
        else:
            config_file = os.path.expanduser(config_file)
        
        self.create_directories(config_dir, config_file)
        with open(config_file, 'r') as f:
            config = yaml.safe_load(f)
        
        if type(config) != dict:
            raise ChisubmitException("Configuration is not valid YAML")

        options = config.keys()

        for p in self.REQUIRED_FIELDS:
            if p not in options:
                raise ChisubmitException("Configuration file does not have required option '%s'" % p)
            options.remove(p)

        for p in options:
            if p not in self.OPTIONAL_FIELDS:
                raise ChisubmitException("Configuration file has invalid option '%s'" % p)

        config['directory'] = config_dir
        self.config_file = config_file
        self.options = config

    def create_directories(self, config_dir, config_file):
        # Create directory if it does not exist
        if not os.path.exists(config_dir):
            os.mkdir(config_dir)

        if not os.path.exists(config_dir + "/courses/"):
            os.mkdir(config_dir + "/courses/")

        if not os.path.exists(config_dir + "/repositories/"):
            os.mkdir(config_dir + "/repositories/")

        if not os.path.exists(config_file):
            example_conf = resource_filename(Requirement.parse("chisubmit"), "chisubmit/config/chisubmit.sample.conf")
            shutil.copyfile(example_conf, config_file)

    def save(self):
        user_options = dict((k, v) for k,v in self.options.items() if k not in self.IMPLICIT_FIELDS)
        try:
            f = open(self.config_file, 'w')
            yaml.safe_dump(user_options, f, default_flow_style=False)
            f.close()
        except IOError, ioe:
            raise ChisubmitException("Error when saving configuration to file %s: %s" % (self.config_file, ioe.meesage), ioe)
        
    def get_server_profile(self, name=None):
        if not self.options.has_key("server"):
            return None
        
        if not self["server"].has_key("profiles"):
            return None
        
        if name is None:
            if not self["server"].has_key("default-profile"):
                return None
            else:
                name = self["server"]["default-profile"]

        if not self["server"]["profiles"].has_key(name):
            return None
        else:
            return self["server"]["profiles"][name]



import abc
from chisubmit.common import ChisubmitException

class ConnectionString(object):

    def __init__(self, s):
        params = s.split(";")
        params = [x.split("=") for x in params]

        try:
            params = dict([(k.strip(), v.strip()) for k,v in params])
        except ValueError, ve:
            raise ChisubmitException("Improperly formatted connection string: %s" % s, original_exception = ve)

        if not params.has_key("server_type"):
            raise ChisubmitException("Connection string does not have 'server_type' parameter: %s" % s)

        self.server_type = params.pop("server_type")
        self.params = params

    @abc.abstractmethod
    def to_string(self):
        pass


class RemoteRepositoryConnectionBase(object):

    def __init__(self, connection_string, staging):
        if connection_string.server_type != self.get_server_type_name():
            raise ChisubmitException("Expected server_type in connection string to be '%s', got '%s'" %
                                     (self.get_server_type_name(), connection_string.server_type))

        param_names = connection_string.params.keys()

        for p in self.get_connstr_mandatory_params():
            if p not in param_names:
                raise ChisubmitException("Connection string does not have required parameter '%s'" % p)
            param_names.remove(p)
            setattr(self, p, connection_string.params[p])

        for p in param_names:
            if p not in self.get_connstr_optional_params():
                raise ChisubmitException("Connection string has invalid parameter '%s'" % p)
            setattr(self, p, connection_string.params[p])

        for p in self.get_connstr_optional_params():
            if p not in param_names:
                setattr(self, p, None)

        self.staging = staging
        self.is_connected = False

    @staticmethod
    @abc.abstractmethod
    def get_server_type_name():
        pass

    @staticmethod
    @abc.abstractmethod
    def get_connstr_mandatory_params():
        pass

    @staticmethod
    @abc.abstractmethod
    def get_connstr_optional_params():
        pass


    @abc.abstractmethod
    def get_credentials(self, username, password, delete_repo = False):
        pass

    @abc.abstractmethod
    def connect(self):
        pass

    @abc.abstractmethod
    def disconnect(self):
        pass

    @abc.abstractmethod
    def deinit_course(self, course):
        pass

    @abc.abstractmethod
    def init_course(self, course, fail_if_exists=True):
        pass

    @abc.abstractmethod
    def update_instructors(self, course):
        pass

    @abc.abstractmethod
    def update_graders(self, course):
        pass

    @abc.abstractmethod
    def create_team_repository(self, course, team, fail_if_exists=True, private=True):
        pass

    @abc.abstractmethod
    def update_team_repository(self, course, team):
        pass

    @abc.abstractmethod
    def exists_team_repository(self, course, team):
        pass

    @abc.abstractmethod
    def get_repository_git_url(self, course, team):
        pass

    @abc.abstractmethod
    def get_repository_http_url(self, course, team):
        pass

    @abc.abstractmethod
    def get_commit(self, course, team, commit_sha):
        pass

    @abc.abstractmethod
    def create_submission_tag(self, course, team, tag_name, tag_message, commit_sha):
        pass

    @abc.abstractmethod
    def update_submission_tag(self, course, team, tag_name, tag_message, commit_sha):
        pass

    @abc.abstractmethod
    def get_submission_tag(self, course, team, tag_name):
        pass

    @abc.abstractmethod
    def delete_team_repository(self, course, team, fail_if_not_exists=True):
        pass
    
    def _get_user_git_username(self, user):
        if self.staging:
            option = "git-staging-username"
        else:
            option = "git-username"
        
        if not hasattr(user, "repo_info") or user.repo_info is None or not user.repo_info.has_key(option): 
            raise ChisubmitException("User '%s' does not have a '%s' value" % (user.user.id, option))
        else:
            return user.repo_info[option]
    
    
    
class GitCommit(object):
    
    def __init__(self, sha, message, 
                 author_name, author_email, authored_date,
                 committer_name, committer_email, committed_date):
        self.sha = sha
        self.message = message
        self.author_name = author_name
        self.author_email = author_email
        self.authored_date = authored_date
        self.committer_name = committer_name
        self.committer_email = committer_email
        self.committed_date = committed_date
        
class GitTag(object):
    
    def __init__(self, name, commit):
        self.name = name
        self.commit = commit



# Needed so we can import from the global "gitlab" package
from __future__ import absolute_import

from gitlab import Gitlab
import gitlab.exceptions

from chisubmit.repos import RemoteRepositoryConnectionBase, GitCommit, GitTag
from chisubmit.common import ChisubmitException
from dateutil.parser import parse

class GitLabConnection(RemoteRepositoryConnectionBase):

    def __init__(self, connection_string, staging):
        RemoteRepositoryConnectionBase.__init__(self, connection_string, staging)
        
        self.gitlab = None
        
        # Map course id's to GitLab group IDs
        self.gitlab_group_id = {}
        
        # Map student id's to GitLab user IDs
        self.gitlab_user_id = {}

        
    @staticmethod
    def get_server_type_name():
        return "GitLab"
    
    @staticmethod
    def get_connstr_mandatory_params():
        return ["gitlab_hostname"]

    @staticmethod
    def get_connstr_optional_params():
        return []
    
    def get_credentials(self, username, password, delete_repo = False):
        try:
            g = Gitlab(self.gitlab_hostname)
            rv = g.login(username, password)
            if not rv:
                return None, False
            else:
                return g.currentuser()["private_token"], True
        except gitlab.exceptions.HttpError, he:
            if he.message == "401 Unauthorized":
                return None, False
            else:
                raise ChisubmitException("Unexpected error getting authorization token (Reason: %s)" % (he.message), he)
    
    def connect(self, credentials):
        # Credentials are a GitLab private token
        self.gitlab = Gitlab(self.gitlab_hostname, token=credentials)    
        try:
            # Test connection by grabbing current user
            user = self.gitlab.currentuser()
            
            if user.has_key("message") and user["message"] == "401 Unauthorized":
                raise ChisubmitException("Invalid GitLab credentials for server '%s'" % (self.gitlab_hostname))
            
            if not user.has_key("username"):
                raise ChisubmitException("Unexpected error connecting to GitLab server '%s'" % (self.gitlab_hostname))
                            
        except Exception as e:
            raise
            raise ChisubmitException("Unexpected error connecting to GitLab server '%s': %s" % (self.gitlab_hostname, e.message))       

    def disconnect(self, credentials):
        pass
    
    def init_course(self, course, fail_if_exists=True):
        group = self.__get_group(course)
        group_name = self.__get_group_name(course)
        if fail_if_exists and group is not None:
            raise ChisubmitException("Course '%s' already has a GitLab group" % group_name)
        
        if group is None:
            if self.staging:
                course_name = course.name + " - STAGING"
            else:
                course_name = course.name

            group_name = self.__get_group_name(course)
            new_group = self.gitlab.creategroup(course_name, group_name)
            if isinstance(new_group, gitlab.exceptions.HttpError):
                print new_group.__dict__
                raise ChisubmitException("Could not create group '%s'" % self.__get_group_name(course), new_group)
                
    def deinit_course(self, course):
        group = self.__get_group(course)
        if group is not None:
            rv = self.gitlab.deletegroup(self.__get_group_id(course))
    
    def update_instructors(self, course):
        for instructor in course.instructors:
            self.__add_user_to_course_group(course, self._get_user_git_username(instructor), "owner")

        # TODO: Remove instructors that may have been removed


    def update_graders(self, course):
        for grader in course.graders:
            self.__add_user_to_course_group(course, self._get_user_git_username(grader), "developer")

        # TODO: Remove instructors that may have been removed

    
    def create_team_repository(self, course, team, fail_if_exists=True, private=True):
        repo_name = self.__get_team_namespaced_project_name(course, team)
        student_names = ", ".join(["%s %s" % (s.user.first_name, s.user.last_name) for s in team.students])
        repo_description = "%s: Team %s (%s)" % (course.name, team.id, student_names)
        
        students = [s for s in course.students if s.user.id in [ts.user.id for ts in team.students]]
        
        if not self.staging:
            gitlab_students = []

            # Make sure users exist
            for s in students:
                gitlab_student = self.__get_user_by_username(self._get_user_git_username(s))
                if gitlab_student is None:
                    raise ChisubmitException("GitLab user '%s' does not exist " % (self._get_user_git_username(s)))
                
                gitlab_students.append(gitlab_student)        

        project = self.__get_team_project(course, team)
        if project is not None and fail_if_exists:
            raise ChisubmitException("Repository %s already exists" % repo_name)
        
        if project is None:
            group = self.__get_group(course)
            
            if group is None:
                raise ChisubmitException("Group for course '%s' does not exist" % course.id)

            # Workaround: Our GitLab server doesn't like public repositories
            #if private:
            #    public = 0
            #else:
            #    public = 1
            
            gitlab_project = self.gitlab.createproject(team.id,
                                                       namespace_id = group["id"],
                                                       description = repo_description,
                                                       public = 0)
            
            if gitlab_project == False:
                raise ChisubmitException("Could not create repository %s" % repo_name)
            
            if not self.staging:                
                for gitlab_student in gitlab_students:
                    rc = self.gitlab.addprojectmember(gitlab_project["id"],
                                                      gitlab_student["id"], 
                                                      "developer")
                    
                    if rc == False:
                        raise ChisubmitException("Unable to add user %s to %s" % (gitlab_student["username"], repo_name))

                                                    
    
    def update_team_repository(self, course, team):
        repo_name = self.__get_team_namespaced_project_name(course, team)

        students = [s for s in course.students if s.user.id in [ts.user.id for ts in team.students]]
        
        gitlab_project = self.__get_team_project(course, team)
        
        for s in students:
            gitlab_student = self.__get_user_by_username(self._get_user_git_username(s))
            if gitlab_student is None:
                raise ChisubmitException("GitLab user '%s' does not exist " % (self._get_user_git_username(s)))
            rc = self.gitlab.addprojectmember(gitlab_project["id"],
                                              gitlab_student["id"], 
                                              "developer")        
            if rc == False:
                raise ChisubmitException("Unable to add user %s to %s" % (gitlab_student["username"], repo_name))
    
    
    def exists_team_repository(self, course, team):
        repo = self.__get_team_project(course, team)
        if repo is None:
            return False
        else:
            return True
    
    def get_repository_git_url(self, course, team):
        repo_name = self.__get_team_namespaced_project_name(course, team)
        return "git@%s:%s.git" % (self.gitlab_hostname, repo_name)
            
    def get_repository_http_url(self, course, team):
        repo_name = self.__get_team_namespaced_project_name(course, team)
        return "https://%s/%s" % (self.gitlab_hostname, repo_name)
    
    def get_commit(self, course, team, commit_sha):
        project_api_id = self.__get_team_project_api_id(course, team)
        gitlab_commit = self.gitlab.getrepositorycommit(project_api_id, commit_sha)
        if gitlab_commit == False:
            return None
        else:
            committer_name = gitlab_commit.get("committer_name", gitlab_commit["author_name"])
            committer_email = gitlab_commit.get("committer_email", gitlab_commit["author_email"])
            
            commit = GitCommit(gitlab_commit["id"], gitlab_commit["title"], 
                 gitlab_commit["author_name"], gitlab_commit["author_email"], parse(gitlab_commit["authored_date"]),
                 committer_name, committer_email, parse(gitlab_commit["committed_date"]))
            return commit
    
    def create_submission_tag(self, course, team, tag_name, tag_message, commit_sha):
        pass 
        # TODO: Commenting out for now, since GitLab doesn't support updating/removing
        #       tags through the API
        #        
        # project_name = self.__get_team_namespaced_project_name(course, team)
        # 
        # commit = self.get_commit(course, team, commit_sha)
        # 
        # if commit is None:
        #     raise ChisubmitException("Cannot create tag %s for commit %s (commit does not exist)" % (tag_name, commit_sha))
        # 
        # rc = self.gitlab.createrepositorytag(project_name, tag_name, commit_sha, tag_message)
        # if rc == False:
        #     raise ChisubmitException("Cannot create tag %s in project %s (error when creating tag)" % (tag_name, project_name))
    
    def update_submission_tag(self, course, team, tag_name, tag_message, commit_sha):
        # TODO: Not currently possible with current GitLab API
        pass
    
    def get_submission_tag(self, course, team, tag_name):
        project_name = self.__get_team_namespaced_project_name(course, team)
        gitlab_tag = self.__get_tag(project_name, tag_name)
        
        if gitlab_tag is None:
            return None
        
        tag = GitTag(name = gitlab_tag["name"],
             commit = self.get_commit(course, team, gitlab_tag["commit"]["id"]))
        
        return tag
    
    def delete_team_repository(self, course, team, fail_if_not_exists):
        project_name = self.__get_team_namespaced_project_name(course, team)
        project_api_id = self.__get_teayJey8z_D6AP-oMaTxu2qm_project_api_id(course, team)
        
        repo = self.__get_team_project(course, team)
        
        if repo is None:
            if fail_if_not_exists:
                raise ChisubmitException("Trying to delete a repository that doesn't exist (%s)" % (project_name))
            else:
                return 
        
        self.gitlab.deleteproject(project_api_id)
    
    def __get_group_name(self, course):
        if self.staging:
            return course.id + "-staging"
        else:
            return course.id    
    
    def __get_group_id(self, course):
        group_name = self.__get_group_name(course)
        if self.gitlab_group_id.has_key(group_name):
            return self.gitlab_group_id[group_name]
        else:
            # TODO: Paginations
            groups = self.gitlab.getgroups()
            
            if groups == False:
                raise ChisubmitException("Unable to fetch Gitlab groups %s (id: %s)" % (self.gitlab_group, self.gitlab_group_id))
    
            for group in groups:
                self.gitlab_group_id[group["path"]] = group["id"]

            if self.gitlab_group_id.has_key(group_name):
                return self.gitlab_group_id[group_name]
            else:
                return None

    def __get_user_by_username(self, username):
        # TODO: Paginations
        users = self.gitlab.getusers(search=username)
        
        if users == False:
            raise ChisubmitException("Unable to fetch Gitlab users")

        if len(users) == 0:
            return None

        for user in users:
            if user["username"] == username:
                return user

        return None

    def __get_user_id(self, user_id):
        if self.gitlab_user_id.has_key(user_id):
            return self.gitlab_user_id[user_id]
        else:
            # TODO: Paginations
            users = self.gitlab.getusers()
            
            if users == False:
                raise ChisubmitException("Unable to fetch Gitlab users")
    
            for user in users:
                self.gitlab_user_id[user_id] = user["id"]

            if self.gitlab_user_id.has_key(user_id):
                return self.gitlab_user_id[user_id]
            else:
                return None
    

    def __get_group(self, course):
        group_id = self.__get_group_id(course)
        
        if group_id == None:
            return None
        
        group = self.gitlab.getgroups(group_id = group_id)
        
        if group == False:
            return None
        else:
            return group
        
    def __get_team_namespaced_project_name(self, course, team):
        group_name = self.__get_group_name(course)
        s = "%s/%s" % (group_name, team.id)
        return s.lower()      
    
    def __get_team_project_api_id(self, course, team):
        project_name = self.__get_team_namespaced_project_name(course, team)
        return project_name.replace("/", "%2F")
    
    def __get_team_project(self, course, team):
        project_api_id = self.__get_team_project_api_id(course, team)
        
        project = self.gitlab.getproject(project_api_id)
        
        if project == False:
            return None
        else:
            return project      
        
    def __add_user_to_course_group(self, course, username, access_level):
        group_name = self.__get_group_name(course)

        user = self.__get_user_by_username(username)
        if user is None:
            raise ChisubmitException("Couldn't add user '%s' to group '%s'. User does not exist" % (username, group_name))
        
        group_id = self.__get_group_id(course)

        if group_id is None:
            raise ChisubmitException("Couldn't add user '%s' to group '%s'. Course group does not exist" % (username, group_name))
                
        self.gitlab.addgroupmember(group_id, user["id"], access_level)
        
        # If the return code is False, we can't distinguish between
        # "failed because the user is already in the group" or
        # "failed for other reason".
        
        # TODO: Check whether user was actually added to group
    
    def __get_tag(self, project_name, tag_name):
        tags = self.gitlab.getrepositorytags(project_name)
        
        if tags == False:
            raise ChisubmitException("Couldn't get tags for project %s" % project_name)
        
        for t in tags:
            if t["id"] == tag_name:
                return t
            
        return None
            
        
    def __has_tag(self, project_name, tag_name):
        return self.__get_tag(project_name, tag_name) is not None            
    
        



from chisubmit.tests.common import ChisubmitTestCase, ChisubmitCLITestClient,\
    cli_test
from chisubmit.tests.fixtures import complete_course
from chisubmit.cli import chisubmit_get_credentials_cmd
from chisubmit.backend.webapp.api.users.models import User
    
class CLIGetChisubmitCredentials(ChisubmitTestCase):
            
    FIXTURE = complete_course
            
    @cli_test
    def test_auth1(self, runner):
        instructor = ChisubmitCLITestClient("instructor1", None, runner,
                                            verbose = True)
        
        user = User.from_id("instructor1")
        user.api_key = None
        self.server.db.session.add(user)
        self.server.db.session.commit()
        
        result = instructor.run(None, 
                                cmd = chisubmit_get_credentials_cmd,
                                cmd_input = "instructor1\n"+self.auth_password+'\n')
        self.assertEquals(result.exit_code, 0)
        
        user = User.from_id("instructor1")
        self.assertIsNotNone(user.api_key)
        self.assertNotEquals(user.api_key, "instructor1")

    @cli_test
    def test_auth2(self, runner):
        instructor = ChisubmitCLITestClient("instructor1", None, runner,
                                            verbose = True)
                
        result = instructor.run(None, 
                                cmd = chisubmit_get_credentials_cmd,
                                cmd_input = "instructor1\n"+self.auth_password+'\n')
        self.assertEquals(result.exit_code, 0)
        
        user = User.from_id("instructor1")
        self.assertIsNotNone(user.api_key)
        self.assertEquals(user.api_key, "instructor1")

    @cli_test
    def test_auth3(self, runner):
        instructor = ChisubmitCLITestClient("instructor1", None, runner,
                                            verbose = True)
                
        result = instructor.run(None, 
                                ["--reset"],
                                cmd = chisubmit_get_credentials_cmd,
                                cmd_input = "instructor1\n"+self.auth_password+'\n')
        self.assertEquals(result.exit_code, 0)
        
        user = User.from_id("instructor1")
        self.assertIsNotNone(user.api_key)
        self.assertNotEquals(user.api_key, "instructor1")




#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import yaml
import textwrap

def feq(a, b, eps=0.01):
    return abs(a - b) <= eps

class ChisubmitRubricException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class RubricFile(object):
    
    FIELD_COMMENTS = "Comments"
    FIELD_PENALTIES = "Penalties"
    FIELD_POINTS   = "Points"
    FIELD_TOTAL_POINTS = "Total Points"
    FIELD_POINTS_POSSIBLE = "Points Possible"
    FIELD_POINTS_OBTAINED = "Points Obtained"
    
    def __init__(self, assignment, points, penalties, comments):
        self.assignment = assignment
        self.points = points
        self.penalties = penalties
        self.comments = comments
        
    def __format_points(self, n):
        return str(round(n,2) if n % 1 else int(n))
        
    def to_yaml(self, include_blank_comments = False):
        # We generate the file manually to make it as human-readable as possible
        
        s = "Points:\n"
        
        total_points_possible = 0
        total_points_obtained = 0
        
        for gc in self.assignment.grade_components:
            s += "%s%s:\n" % (" "*4, gc.description)
            s += "%s%s: %s\n" % (" "*8, self.FIELD_POINTS_POSSIBLE, self.__format_points(gc.points))
            total_points_possible += gc.points
            if self.points[gc.description] is None:
                p = ""
            else:
                total_points_obtained += self.points[gc.description]
                p = self.__format_points(self.points[gc.description])
                
            s += "%s%s: %s\n" % (" "*8, self.FIELD_POINTS_OBTAINED, p)
            s += "\n" 
            
        penalty_points = 0.0
        if self.penalties is not None:
            s += "%s:\n" % self.FIELD_PENALTIES
            for desc, v in self.penalties.items():
                penalty_points += v
                s += "%s%s:%s\n" % (" "*4, desc, self.__format_points(v))            
            s += "\n"

        s += "%s: %s / %s\n" % (self.FIELD_TOTAL_POINTS,
                              self.__format_points(total_points_obtained + penalty_points),
                              self.__format_points(total_points_possible)) 
            
        if self.comments is not None or include_blank_comments:
            s += "\n"
            s += "%s: >\n" % self.FIELD_COMMENTS
            
            if self.comments is not None:
                for l in self.comments.strip().split("\n"):
                    for l2 in textwrap.wrap(l, initial_indent = " "*4):
                        s += l2 + "\n" 
            else:
                s += " "*4 + "None"
        
        return s
        
    def save(self, rubric_file, include_blank_comments = False):
        try:
            f = open(rubric_file, 'w')
            f.write(self.to_yaml(include_blank_comments))
            f.close()
        except IOError, ioe:
            raise ChisubmitRubricException("Error when saving rubric to file %s: %s" % (rubric_file, ioe.message), ioe)
        
    @classmethod
    def from_file(cls, rubric_file, assignment):
        rubric = yaml.load(rubric_file)

        if not rubric.has_key(RubricFile.FIELD_POINTS):
            raise ChisubmitRubricException("Rubric file doesn't have a '%s' field." % RubricFile.FIELD_POINTS)

        if not rubric.has_key(RubricFile.FIELD_TOTAL_POINTS):
            raise ChisubmitRubricException("Rubric file doesn't have a '%s' field." % RubricFile.FIELD_TOTAL_POINTS)

        points = {}
        total_points_obtained = 0
        total_points_possible = 0
        for grade_component in assignment.grade_components:
            if not rubric[RubricFile.FIELD_POINTS].has_key(grade_component.description):
                raise ChisubmitRubricException("Rubric is missing '%s' points." % grade_component.description)
            
            component = rubric[RubricFile.FIELD_POINTS][grade_component.description]
            
            if not component.has_key(RubricFile.FIELD_POINTS_POSSIBLE):
                raise ChisubmitRubricException("Grade component '%s' is missing '%s' field." % (grade_component.description, RubricFile.FIELD_POINTS_POSSIBLE))

            if not component.has_key(RubricFile.FIELD_POINTS_OBTAINED):
                raise ChisubmitRubricException("Grade component '%s' is missing '%s' field." % (grade_component.description, RubricFile.FIELD_POINTS_OBTAINED))
            
            points_possible = component[RubricFile.FIELD_POINTS_POSSIBLE]
            points_obtained = component[RubricFile.FIELD_POINTS_OBTAINED]
            
            if points_possible != grade_component.points:
                raise ChisubmitRubricException("Grade component '%s' in rubric has incorrect possible points (expected %i, got %i)" %
                                                (grade_component.description, grade_component.points, points_possible))
                
            if points_obtained is not None:
                if points_obtained < 0:
                    raise ChisubmitRubricException("Grade component '%s' in rubric has negative points (%i)" %
                                                    (grade_component.description, points_obtained))
    
                if points_obtained > points_possible:
                    raise ChisubmitRubricException("Grade component '%s' in rubric has more than allowed points (%i > %i)" %
                                                    (grade_component.description, points_obtained, points_possible))

                total_points_obtained += points_obtained

            points[grade_component.description] = points_obtained
            total_points_possible += grade_component.points

        penalty_points = 0.0
        if rubric.has_key(RubricFile.FIELD_PENALTIES):
            penalties = rubric[RubricFile.FIELD_PENALTIES]
            for desc, v in penalties.items():
                if v >= 0:
                    raise ChisubmitRubricException("Rubric file has a non-negative penalty: %s (%s)" % (v, desc))
                penalty_points += v
        else:
            penalties = None

        if type(rubric[RubricFile.FIELD_TOTAL_POINTS]) != str:
            raise ChisubmitRubricException("Total points is not a string: %s" % rubric[RubricFile.FIELD_TOTAL_POINTS])
        
        total_points = rubric[RubricFile.FIELD_TOTAL_POINTS].split(" / ")
        if len(total_points) != 2:
            raise ChisubmitRubricException("Improperly formatted total points: %s" % rubric[RubricFile.FIELD_TOTAL_POINTS])
        
        if not feq(float(total_points[0]), float(total_points_obtained) + penalty_points):
            raise ChisubmitRubricException("Incorrect number of total points obtained (Expected %.2f, got %.2f)" % 
                                           (float(total_points_obtained) + penalty_points, float(total_points[0])))
            
        if not feq(float(total_points[1]), float(total_points_possible)):
            raise ChisubmitRubricException("Incorrect number of total points obtained (Expected %.2f, got %.2f)" % 
                                           (float(total_points_possible), float(total_points[1])))
            
        if not rubric.has_key(RubricFile.FIELD_COMMENTS):
            comments = None
        else:
            comments = rubric[RubricFile.FIELD_COMMENTS]

        return cls(assignment, points, penalties, comments)

    @classmethod
    def from_assignment(cls, assignment, team_assignment = None):
        gcs = dict([(gc.id, gc.description) for gc in assignment.grade_components])
        points = dict([(gc.description, None) for gc in assignment.grade_components])
        
        if team_assignment is not None:
            for grade in team_assignment.grades:
                gc_id = grade.grade_component_id
                points[gcs[gc_id]] = grade.points
        
        return cls(assignment, points, penalties = None, comments = None)




#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import datetime
import pytz
import random
import hashlib
import base64
import string
from tzlocal import get_localzone
from chisubmit.repos.factory import RemoteRepositoryConnectionFactory
import math

localzone = get_localzone()

now_override = None

def get_datetime_now_utc():
    global now_override
    
    if now_override is None:
        return datetime.datetime.now(pytz.utc).replace(microsecond=0)
    else:
        return now_override 
    

def set_testing_now(dt):
    global now_override
    
    now_override = dt
    

def set_datetime_timezone_utc(dt):
    return pytz.utc.localize(dt)

def convert_datetime_to_utc(dt, default_tz = localzone):
    if dt.tzinfo is None:
        dt = localzone.localize(dt)
    return dt.astimezone(pytz.utc)

def convert_datetime_to_local(dt, default_tz = localzone):
    if dt.tzinfo is None:
        dt = localzone.localize(dt)
    return dt.astimezone(localzone)

def compute_extensions_needed(submission_time, deadline):
    delta = (submission_time - deadline).total_seconds()

    extensions_needed = math.ceil(delta / (3600.0 * 24))

    if extensions_needed <= 0:
        return 0
    else:
        return int(extensions_needed)

# Based on http://jetfar.com/simple-api-key-generation-in-python/
def gen_api_key():
    s = str(random.getrandbits(256))
    h = hashlib.sha256(s)
    altchars = random.choice(string.ascii_letters) + random.choice(string.ascii_letters)
    b = base64.b64encode(h.digest(), altchars).rstrip("==")
    return unicode(b)
    
    
def create_connection(course, config, staging = False):
    if not staging:
        connstr = course.options["git-server-connstr"]
    else:
        connstr = course.options["git-staging-connstr"]

    conn = RemoteRepositoryConnectionFactory.create_connection(connstr, staging)
    server_type = conn.get_server_type_name()
    
    git_credentials = None
    if config['git-credentials'] is not None:
        git_credentials = config['git-credentials'].get(server_type, None)

    if git_credentials is None:
        print "You do not have %s credentials." % server_type
        return None
    else:
        conn.connect(git_credentials)
        return conn
           

    
    



from flask import jsonify, request, abort
from chisubmit.backend.webapp.api import db, app
from chisubmit.backend.webapp.api.blueprints import api_endpoint
from chisubmit.backend.webapp.api.users.models import User
from chisubmit.backend.webapp.api.users.forms import CreateUserInput, UpdateUserInput,\
    GenerateAccessTokenInput
from chisubmit.backend.webapp.auth import ldap, require_auth
from chisubmit.backend.webapp.auth.token import require_apikey
from chisubmit.backend.webapp.auth.authz import require_admin_access
from chisubmit.common.utils import gen_api_key
from flask.globals import g


@api_endpoint.route('/users', methods=['POST'])
@require_apikey
@require_admin_access
def users():
    input_data = request.get_json(force=True)
    if not isinstance(input_data, dict):
        return jsonify(error='Request data must be a JSON Object'), 400

    form = CreateUserInput.from_json(input_data)
    if not form.validate():
        return jsonify(errors=form.errors), 400

    user = User()
    form.populate_obj(user)
    db.session.add(user)
    db.session.commit()

    return jsonify({'user': user.to_dict()}), 201


@api_endpoint.route('/users/<user_id>', methods=['PUT', 'GET'])
@require_apikey
@require_admin_access
def user(user_id):
    user = User.query.filter_by(id=user_id).first()
    # TODO 11DEC14: check permissions *before* 404
    if user is None:
        abort(404)

    if request.method == 'PUT':
        input_data = request.get_json(force=True)
        if not isinstance(input_data, dict):
            return jsonify(error='Request data must be a JSON Object'), 400
        form = UpdateUserInput.from_json(input_data)
        if not form.validate():
            return jsonify(errors=form.errors), 400

        user.set_columns(**form.patch_data)
        db.session.commit()

    return jsonify({'user': user.to_dict()})


@api_endpoint.route('/auth', methods=['POST'])
@require_auth
def get_api_key():
    input_data = request.get_json(force=True)
    if not isinstance(input_data, dict):
        return jsonify(error='Request data must be a JSON Object'), 400
    form = GenerateAccessTokenInput.from_json(input_data)
    if not form.validate():
        return jsonify(errors=form.errors), 400
    
    exists_prior = g.user.api_key is not None 

    if form.reset.data or g.user.api_key is None:
        api_key = gen_api_key()
        g.user.api_key = api_key
        db.session.add(g.user)
        db.session.commit()
        is_new = True
    else:
        api_key = g.user.api_key
        is_new = False
        
    return jsonify({'api_key': api_key, 
                    'exists_prior': exists_prior, 
                    'is_new': is_new})



from chisubmit.tests.common import ChisubmitTestCase, ChisubmitMultiTestCase
from chisubmit.tests.fixtures import complete_course, users_and_courses
import json

class Empty(ChisubmitTestCase):
    
    def test_get_courses(self):
        c = self.get_admin_test_client()
        
        response = c.get("courses")
        self.assert_http_code(response, 200)
        
        data = json.loads(response.get_data())        
        self.assertIn("courses", data)
        self.assertEquals(len(data["courses"]), 0)
        
    def test_create_course(self):
        c = self.get_admin_test_client()
        
        data = {"id": "cmsc12300", "name": "Foobarmentals of Foobar"}        
        response = c.post("courses", data)
        self.assert_http_code(response, 201)
           

class CoursesAndUsers(ChisubmitMultiTestCase):
    
    FIXTURE = users_and_courses
    
    def __get_courses(self, client, expected):
        response = client.get("courses")
        self.assert_http_code(response, 200)
        
        data = json.loads(response.get_data())        
        self.assertIn("courses", data)
        self.assertEquals(len(data["courses"]), expected)
        
        return data

    def __get_course(self, client, course):
        response = client.get("courses/%s" % course["id"])
        self.assert_http_code(response, 200)
        
        data = json.loads(response.get_data())        
        self.assertIn("course", data)
        self.assertEquals(len(data), 1)
        self.assertEquals(data["course"]["name"], course["name"])
        
        return data

    
    def test_get_courses_admin(self):
        c = self.get_admin_test_client()
        data = self.__get_courses(c, len(self.FIXTURE["courses"]))
        
    def test_get_courses(self):
        c = self.get_test_client(self.FIXTURE["users"]["instructor1"])
        data = self.__get_courses(c, 1)
           
        c = self.get_test_client(self.FIXTURE["users"]["grader1"])        
        data = self.__get_courses(c, 1)

        c = self.get_test_client(self.FIXTURE["users"]["student1"])        
        data = self.__get_courses(c, 1)


    def test_get_course_admin(self):
        c = self.get_admin_test_client()
        data = self.__get_course(c, self.FIXTURE["courses"]["cmsc40100"])
        
    def test_get_course(self):
        c = self.get_test_client(self.FIXTURE["users"]["instructor1"])
        data = self.__get_course(c, self.FIXTURE["courses"]["cmsc40100"])

        c = self.get_test_client(self.FIXTURE["users"]["grader1"])        
        data = self.__get_course(c, self.FIXTURE["courses"]["cmsc40100"])

        c = self.get_test_client(self.FIXTURE["users"]["student1"])        
        data = self.__get_course(c, self.FIXTURE["courses"]["cmsc40100"])
         
class CompleteCourse(ChisubmitTestCase):
    
    FIXTURE = complete_course
        
    def test_update_student1(self):
        from chisubmit.backend.webapp.api.courses.models import CoursesStudents

        c = self.get_admin_test_client()
        
        attrs = {'name': 'git_username', 'value': 'foobar'}
        
        data = {"students": 
                {
                 "update": [
                            {"student_id": "student1",
                             "repo_info": [attrs]}
                           ]
                }
               }
                                     
        response = c.put("courses/cmsc40100", data)
        
        course_student = CoursesStudents.query.filter_by(
                        course_id="cmsc40100").filter_by(
                        student_id="student1").first()
                        

    def test_update_student2(self):
        from chisubmit.backend.webapp.api.courses.models import CoursesStudents

        c = self.get_admin_test_client()
        
        
        data = {"students": 
                {
                 "update": [
                            {"student_id": "student1",
                             "dropped": True}
                           ]
                }
               }
                                     
        response = c.put("courses/cmsc40100", data)
        
        course_student = CoursesStudents.query.filter_by(
                        course_id="cmsc40100").filter_by(
                        student_id="student1").first()
                        
                


from wtforms import Form
from wtforms.validators import InputRequired, Optional, Length
from wtforms.fields import StringField, IntegerField, FormField, FieldList
from chisubmit.backend.webapp.api.forms import ISODateTimeField
from wtforms.fields.core import BooleanField, FloatField


class CreateGradeComponentInput(Form):
    id = StringField(validators=[Length(max=36), InputRequired()])
    description = StringField(validators=[InputRequired()])
    points = FloatField(validators=[InputRequired()])


class AddGradeComponents(Form):
    add = FieldList(FormField(CreateGradeComponentInput))


class UpdateAssignmentInput(Form):
    id = StringField(validators=[Length(max=36, min=2), Optional()])
    name = StringField(validators=[Length(max=36, min=2), Optional()])
    deadline = ISODateTimeField(validators=[Optional()])
    grade_components = FormField(AddGradeComponents)


class CreateAssignmentInput(Form):
    id = StringField(validators=[Length(max=36), InputRequired()])
    deadline = ISODateTimeField(validators=[InputRequired()])
    name = StringField('name',
                       validators=[Length(max=36, min=2),
                                   InputRequired()])


class SearchAssignmentInput(Form):
    assignment_id = StringField('assignment_id',
                             validators=[Length(max=36, min=2),
                                         InputRequired()])
    name = StringField('name', validators=[Length(max=36, min=2), Optional()])


class RegisterAssignmentInput(Form):
    team_name = StringField(validators=[Length(max=36, min=2), Optional()])
    partners = FieldList(StringField())
    
class SubmitAssignmentInput(Form):
    team_id = StringField(validators=[Length(max=36, min=2),InputRequired()])
    commit_sha = StringField(validators=[Length(max=40, min=20),InputRequired()])
    extensions = IntegerField(default=0)
    dry_run = BooleanField(default=False)
        
class CancelSubmitAssignmentInput(Form):
    team_id = StringField(validators=[Length(max=36, min=2),InputRequired()])





#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import logging

def init_logging(verbose = False, debug = False):
    # debug implies verbose
    if verbose:
        level = logging.INFO
    elif debug:
        level = logging.DEBUG
    else:
        level = logging.WARNING
    
    l = logging.getLogger("chisubmit")
    l.setLevel(logging.DEBUG)
    
    fh = logging.StreamHandler()
    fh.setLevel(level)
    formatter = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s')
    fh.setFormatter(formatter)
    l.addHandler(fh)        

def log(msg, func):
    func(msg)

def debug(msg):
    log(msg, logging.getLogger('chisubmit').debug)

def warning(msg):
    log(msg, logging.getLogger('chisubmit').warning)
    
def info(msg):
    log(msg, logging.getLogger('chisubmit').info)



#!/usr/bin/env python

from flask.ext.script import Manager
from flask import url_for
from chisubmit.backend.webapp.api import app

manager = Manager(app)


@manager.command
def routes():
    import urllib
    output = []
    for rule in app.url_map.iter_rules():

        options = {}
        for arg in rule.arguments:
            options[arg] = "[{0}]".format(arg)

        methods = ','.join(rule.methods)
        url = url_for(rule.endpoint, **options)
        line = urllib.unquote("{:50s} {:20s} {}".format(rule.endpoint,
                              methods, url))
        output.append(line)

    for line in sorted(output):
        print line

if __name__ == "__main__":
    manager.run()



#!/usr/bin/env python
# -------------------------------------------------------------------------- #
# Copyright (c) 2013-2014, The University of Chicago
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# - Neither the name of The University of Chicago nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# -------------------------------------------------------------------------- #

from ez_setup import use_setuptools
use_setuptools(version="3.1")
from setuptools import setup, find_packages
import sys

sys.path.insert(0, './src')
from chisubmit import RELEASE


eps = ['chisubmit = chisubmit.cli:chisubmit_cmd_wrapper',
       'chisubmit-get-credentials = chisubmit.cli:chisubmit_get_credentials_cmd_wrapper',
       'chisubmit-server = chisubmit.cli:chisubmit_server_cmd [server]']

setup(name='chisubmit',
      version=RELEASE,
      description='A system for managing assignments and teams in university courses',
      author='University of Chicago, Department of Computer Science',
      author_email='borja@cs.uchicago.edu',
      url='http://www.cs.uchicago.edu/',
      package_dir = {'': 'src'},      
      package_data = {'': ['src/config/*.conf']},
      packages=find_packages("src"),
      
      install_requires = [ "PyGithub >= 1.25.2", "pyapi-gitlab >= 7.5", "GitPython >= 0.3.5",
                           "pyyaml >= 3.10", "pytz >= 2013.8", "tzlocal >= 1.0", "python-dateutil >= 2.4",
                           "click >= 3.3", "requests >= 2.5.1",
                           "colorama >= 0.3.2", "docutils >= 0.12" ],
      extras_require = {
                         "server" : ["flask >= 0.10.1", "flask-sqlalchemy >= 2.0",
                                     "flask-restless >= 0.15", "Flask-Script >= 2.0.5", 
                                     "python-ldap >= 2.4.18", "WTForms-JSON >= 0.2.8"] 
                        },
      setup_requires = [ "setuptools_git >= 1.0" ],
      include_package_data=True,
      
      entry_points = {
        'console_scripts': eps
        },

      zip_safe = False,

      license="Apache Software License",
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Education',
          'License :: OSI Approved :: BSD License',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Topic :: Education'
          ]
     )



from chisubmit.tests.common import cli_test, ChisubmitIntegrationTestCase
from chisubmit.common.utils import get_datetime_now_utc, set_testing_now
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL

from datetime import timedelta
import time

class CLICompleteWorkflowCancelSubmission(ChisubmitIntegrationTestCase):
            
    @cli_test
    def test_complete_with_submission_cancelling(self, runner):
        course_id = u"cmsc40300"
        course_name = u"Foobarmentals of Foobar II"

        admin_id = u"admin"
        instructor_ids = [u"instructor"]
        grader_ids= [u"grader"]
        student_ids = [u"student1", u"student2", u"student3", u"student4"]
        
        all_users = instructor_ids + grader_ids + student_ids
        
        admin, instructors, graders, students = self.create_clients(runner, course_id, admin_id, instructor_ids, grader_ids, student_ids)
        self.create_users(admin, all_users)
        
        self.create_course(admin, course_id, course_name)

        result = admin.run("admin course set-option %s default-extensions 3" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        result = admin.run("admin course set-option %s extension-policy per_student" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        self.add_users_to_course(admin, course_id, instructors, graders, students)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23)
        deadline = deadline.isoformat(sep=" ")

        result = instructors[0].run("instructor assignment add", 
                                    ["pa1", "Programming Assignment 1", deadline])
        self.assertEquals(result.exit_code, 0)
        
        teams = [u"the-flaming-foobars", 
                 u"the-magnificent-mallocs"]        

        students_team = [ (students[0], students[1]),
                          (students[2], students[3])]
        

        self.register_team(students_team[0], teams[0], "pa1", course_id)
        self.register_team(students_team[1], teams[1], "pa1", course_id)
        
        _, _, team_commits = self.create_team_repos(admin, course_id, teams[0:2], students_team[0:2])
        

        # Team 0 cancels their submission
        # Fails because there is nothing to cancel
        result = students_team[0][0].run("student assignment cancel-submit", 
                                         [teams[0], "pa1", "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_FAIL)           

        
        # Team 0 and 1 submit with one extension to pa1
        result = students_team[0][0].run("student assignment submit", 
                                         [teams[0], "pa1", team_commits[0][0].hexsha, 
                                          "--extensions", "1",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
        
        result = students_team[1][0].run("student assignment submit", 
                                         [teams[1], "pa1", team_commits[1][0].hexsha, 
                                          "--extensions", "1",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)


        # Team 0 cancels their submission
        result = students_team[0][0].run("student assignment cancel-submit", 
                                         [teams[0], "pa1", "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        
        # Team 0 cancels their submission (again)
        # Fails because there is nothing to cancel
        result = students_team[0][0].run("student assignment cancel-submit", 
                                         [teams[0], "pa1", "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_FAIL)        


        # Team 1 resubmits and is successful because the deadline hasn't passed yet
        result = students_team[1][0].run("student assignment submit", 
                                         [teams[1], "pa1", team_commits[1][1].hexsha, 
                                          "--extensions", "1",
                                          "--force",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        
        # Let the deadline "pass"
        new_now = get_datetime_now_utc() + timedelta(hours=2)
        set_testing_now(new_now)

        print
        print "~~~ Time has moved 'forward' by two hours ~~~"
        print
        
        # Team 0 submits and is successful because they cancelled their previous submission
        result = students_team[0][0].run("student assignment submit", 
                                         [teams[0], "pa1", team_commits[0][1].hexsha, 
                                          "--extensions", "2",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)


        # Team 1 submits and fails because their previous submission is final
        result = students_team[1][0].run("student assignment submit", 
                                         [teams[1], "pa1", team_commits[1][0].hexsha, 
                                          "--extensions", "2",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_FAIL)
        
        # Team 1 cancels their submission
        # Fails because the previous submission is final
        result = students_team[1][0].run("student assignment cancel-submit", 
                                         [teams[1], "pa1", "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_FAIL)        
 

        for team, student_team in zip(teams, students_team):
            result = student_team[0].run("student team show", [team])
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        
        



import click
from chisubmit.cli.instructor.assignment import instructor_assignment
from chisubmit.cli.instructor.team import instructor_team
from chisubmit.cli.instructor.user import instructor_user
from chisubmit.cli.instructor.course import instructor_course
from chisubmit.cli.instructor.grading import instructor_grading

@click.group()
@click.pass_context
def instructor(ctx):
    pass

instructor.add_command(instructor_assignment)
instructor.add_command(instructor_grading)
instructor.add_command(instructor_team)
instructor.add_command(instructor_user)
instructor.add_command(instructor_course)


from . import test_courses, test_fixtures
from unittest.suite import TestSuite

test_cases = (test_courses, test_fixtures)

def load_tests(loader, tests, pattern):
    suite = TestSuite()
    for test_module in test_cases:
        tests = loader.loadTestsFromModule(test_module)
        suite.addTests(tests)
    return suite


import click
from chisubmit.cli.common import pass_course
import operator
from chisubmit.common.utils import convert_datetime_to_local
from chisubmit.common import CHISUBMIT_SUCCESS


@click.command(name="list")
@click.option('--ids', is_flag=True)
@click.option('--utc', is_flag=True)
@pass_course
@click.pass_context
def shared_assignment_list(ctx, course, ids, utc):
    assignments = course.assignments[:]
    assignments.sort(key=operator.attrgetter("deadline"))

    for assignment in assignments:
        if ids:
            print assignment.id
        else:
            if utc:
                deadline = assignment.deadline.isoformat(" ")
            else:
                deadline = convert_datetime_to_local(assignment.deadline).isoformat(" ")

            fields = [assignment.id, deadline, assignment.name]

            print "\t".join(fields)

    return CHISUBMIT_SUCCESS


import os.path

from chisubmit.common import ChisubmitException
from chisubmit.common.utils import gen_api_key
from chisubmit.backend.webapp.api.json_encoder import CustomJSONEncoder
from chisubmit.backend.webapp.api.blueprints import api_endpoint

from flask.app import Flask
from flask.ext.sqlalchemy import SQLAlchemy

import wtforms_json
from chisubmit.backend.webapp.auth.testing import TestingAuth
from chisubmit.backend.webapp.auth import set_auth
from chisubmit.backend.webapp.auth.ldap import LDAPAuth

wtforms_json.init()

app = Flask(__name__)
app.json_encoder = CustomJSONEncoder

db = SQLAlchemy(app)

import chisubmit.backend.webapp.api.views
app.register_blueprint(api_endpoint, url_prefix='/api/v0')


class ChisubmitAPIServer(object):

    def __init__(self, debug):
        self.app = app        
        self.db = db
        self.debug = debug
                    
        if self.debug:
            self.debug = True
            self.app.config["DEBUG"] = True
            self.app.config["TESTING"] = True
            self.app.config["SQLALCHEMY_RECORD_QUERIES"] = True
        else:
            self.debug = False
            
            
    def connect_sqlite(self, dbfile):
        self.app.config["SQLALCHEMY_DATABASE_URI"] = 'sqlite:///' + dbfile
        
    def connect_postgres(self):
        pass
    
    def set_auth_testing(self, password):
        auth = TestingAuth(self, password)
        set_auth(auth)
    
    def set_auth_ldap(self, ldap_server_uri, base_dn):
        auth = LDAPAuth(self, ldap_server_uri, base_dn)
        set_auth(auth)

    def start(self):
        self.app.run(debug = self.debug)
        
    def init_db(self, drop_all=False):
        if drop_all:
            self.db.drop_all()
        self.db.create_all()

    def create_admin(self, api_key=None):
        from chisubmit.backend.webapp.api.users.models import User

        admin = User.query.filter_by(id=u"admin").first()
        
        if admin is not None:
            return None
        
        if api_key is None:
            api_key = gen_api_key()
            
        admin = User(first_name=u"Administrator", 
                       last_name=u"Administrator", 
                       id=u"admin",
                       api_key=unicode(api_key), 
                       admin=1)
        
        self.db.session.add(admin)
        self.db.session.commit()
        
        return api_key
            
        
        




import git
from git.exc import GitCommandError
from chisubmit.common import ChisubmitException
from gitdb.exc import BadObject
from chisubmit.repos import GitCommit, GitTag
import datetime

class LocalGitRepo(object):
    def __init__(self, directory):
        self.repo = git.Repo(directory)

        self.remotes = dict([(r.name, r) for r in self.repo.remotes])

    @classmethod
    def create_repo(cls, directory, bare = False, clone_from_url = None, remotes = []):
        if clone_from_url is None:
            git.Repo.init(directory, bare = bare)
        else:
            repo = git.Repo.clone_from(clone_from_url, directory)

            for remote_name, remote_url in remotes:
                repo.create_remote(remote_name, remote_url)

            return cls(directory)

    def fetch(self, remote_name, branch = None):
        if branch is None:
            self.remotes[remote_name].fetch()
        else:
            self.remotes[remote_name].fetch("%s:%s" % (branch, branch))

    def reset_branch(self, remote_name, branch):
        branch_refpath = "refs/heads/%s" % branch
        remote_branch_refpath = "refs/remotes/%s/%s" % (remote_name, branch)

        branch_head = self.__get_head(branch_refpath)
        remote_branch = self.__get_ref(remote_branch_refpath)

        if branch_head is None:
            raise ChisubmitException("No such branch: %s" % branch)

        if remote_branch is None:
            raise ChisubmitException("No such remote branch: %s" % branch)

        if self.repo.head.is_detached or self.repo.head.ref != branch_head:
            try:
                branch_head.checkout()
            except GitCommandError, gce:
                print gce
                raise ChisubmitException("Error checking out")

        self.repo.head.reset(remote_branch.commit, index=True, working_tree=True)

    def has_branch(self, branch):
        return (self.__get_branch(branch) is not None)

    def has_remote_branch(self, remote_name, branch):
        return (self.__get_ref("refs/remotes/%s/%s" % (remote_name, branch)) is not None)

    def checkout_branch(self, branch):
        branch_refpath = "refs/heads/%s" % branch
        branch_head = self.__get_head(branch_refpath)

        if branch_head is None:
            raise ChisubmitException("No such branch: %s" % branch)

        try:
            branch_head.checkout()
        except GitCommandError:
            raise ChisubmitException("Error checking out")

    def checkout_commit(self, commit_sha):
        commit = self.get_commit(commit_sha)
        
        if commit is None:
            raise ChisubmitException("Tried to checkout a commit that does not exist: %s" % commit_sha)
        
        try:
            self.repo.git.checkout(commit_sha)
        except GitCommandError:
            raise ChisubmitException("Could not checkout commit %s" % commit_sha)
        

    def get_commit(self, commit_sha):
        try:
            commit = self.repo.commit(commit_sha)
            commit_obj = self.__create_commit_object(commit)
            return commit_obj
        except BadObject, bo:
            return None
        

    def get_tag(self, tag):
        tags = [t for t in self.repo.tags if t.name == tag]

        if len(tags) == 0:
            return None
        else:
            return self.__create_tag_object(tags[0])

    def has_tag(self, tag):
        return (self.get_tag(tag) is not None)

    def create_tag(self, tag_name, commit_sha, message, force = False):
        self.repo.create_tag(tag_name, commit_sha, message, force)

    def create_branch(self, branch, commit):
        self.repo.create_head(branch, commit)

    def push(self, remote_name, branch):
        self.remotes[remote_name].push("%s:%s" % (branch, branch))

    def pull(self, remote_name, branch):
        self.remotes[remote_name].pull("%s" % (branch))
        
    def commit(self, files, commit_message):
        self.repo.index.add(files)
        if self.repo.is_dirty():
            self.repo.index.commit(commit_message)
            return True
        else:
            return False       
    
    def is_dirty(self):
        return self.repo.is_dirty() 

    def __create_tag_object(self, tag):
        return GitTag(name = tag.name,
                      commit = self.__create_commit_object(tag.commit))            


    def __create_commit_object(self, commit):
        authored_date = datetime.datetime.fromtimestamp(commit.authored_date)
        committed_date = datetime.datetime.fromtimestamp(commit.committed_date)
        
        return GitCommit(commit.hexsha, commit.message,
                         commit.author.name, commit.author.email, authored_date, 
                         commit.committer.name, commit.committer.email, committed_date)            
        

    def __get_head(self, path):
        heads = [h for h in self.repo.heads if h.path == path]

        if len(heads) == 0:
            return None
        else:
            return heads[0]

    def __get_branch(self, branch):
        branches = [b for b in self.repo.branches if b.name == branch]

        if len(branches) == 0:
            return None
        else:
            return branches[0]

    def __get_ref(self, path):
        refs = [r for r in self.repo.refs if r.path == path]

        if len(refs) == 0:
            return None
        else:
            return refs[0]



from functools import wraps
from flask import request, abort, g
from chisubmit.backend.webapp.api.users.models import User


def require_admin_access(view_function):
    @wraps(view_function)
    
    # the new, post-decoration function. Note *args and **kwargs here.
    def decorated_function(*args, **kwargs):
        if g.user.admin:
            return view_function(*args, **kwargs)
        else:
            abort(403)

    return decorated_function

def check_roles(user, course, roles):
    if roles is None:
        return True
    else:
        if "student" in roles and user.is_student_in(course):
            return True
        elif "instructor" in roles and user.is_instructor_in(course):
            return True
        elif "grader" in roles and user.is_grader_in(course):
            return True
        else:
            return False
        

def check_admin_access_or_abort(user, status_code):
    if user.admin:
        return
    else:
        abort(status_code)

def check_user_ids_equal_or_abort(user1_id, user2_id, status_code):
    if user1_id == user2_id:
        return
    else:
        abort(status_code)

def check_course_access_or_abort(user, course, status_code, roles = None):
    if user.admin:
        return
    
    if not check_roles(user, course, roles):
        abort(status_code)
    
    if user.is_in_course(course):
        return
    else:
        abort(status_code)
        
def check_team_access_or_abort(user, team, status_code, roles = None):
    if user.admin:
        return
    
    course = team.course

    if not check_roles(user, course, roles):
        abort(status_code)
    
    if user.is_in_course(course):
        if user.is_instructor_in(course) or user.is_grader_in(course):
            return
        elif user.is_student_in(course) and user in team.students:
            return
        else:
            abort(status_code)
    else:
        abort(status_code)        
        